from pydantic import BaseModel


class FileDescriptor(BaseModel):

    f_name: str
    f_url: str
    image: str
    dtc: str
    dtm: str
    size: str
    metadata: dict[str, str]


class FieldSpec(BaseModel):
    id: str
    name: str
    description: str
    dtype: str


class StreamSpec(BaseModel):

    id: str
    name: str
    description: str
    unit: str
    frequency: int
    fields: list[FieldSpec]
    index: list[FieldSpec]



import logging
import os
import re
import tarfile
import zipfile
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path
from urllib.parse import unquote

import parse
from flask import Response, request
from hurry.filesize import size

from .typing import FileDescriptor

Chunk = tuple[bytes, int, int, int]

logger = logging.getLogger(__name__)


def make_tarfile(output_filename: Path, source_dir: Path, dest_dir: str):
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname='.', recursive=True)


def get_chunk(full_path: Path, start_byte: int, end_byte: int | None = None) -> Chunk:
    """ """
    file_size = full_path.stat().st_size
    if end_byte:
        length = end_byte + 1 - start_byte
    else:
        length = file_size - start_byte
    with open(full_path, "rb") as f:
        f.seek(start_byte)
        chunk = f.read(length)
    return chunk, start_byte, length, file_size


def send_media(file_path: Path, mimetype: str) -> Response:
    """ """
    range_header = request.headers.get("Range", None)
    start_byte, end_byte = 0, None
    if range_header:
        match = re.search(r"(\d+)-(\d*)", range_header)
        assert match
        groups = match.groups()
        if groups[0]:
            start_byte = int(groups[0])
        if groups[1]:
            end_byte = int(groups[1])

    chunk, start, length, file_size = get_chunk(file_path, start_byte, end_byte)
    resp = Response(chunk, 206, mimetype=f"video/{mimetype}", content_type=mimetype, direct_passthrough=True)
    resp.headers.add("Content-Range", "bytes {0}-{1}/{2}".format(start, start + length - 1, file_size))
    return resp


def get_file_extension(fp: Path) -> str:
    """ """
    return fp.suffix.lower()


def is_media(fp: Path, ext_dict: dict) -> tuple[bool, str, str]:
    """ """
    tp = "other"
    ext = get_file_extension(fp)
    if ext in ext_dict:
        tp = ext_dict[ext][0]
    return tp in ["audio", "video"], tp, ext


def get_icon(fp: Path, ext_dict: dict) -> str:
    if fp.is_dir():
        return "/static/icons/folder.png"
    if fp.is_file():
        ext = get_file_extension(fp)
        if ext in ext_dict:
            return ext_dict[ext][1]
    return "/static/icons/file.png"


def zip_directory(dest_path: Path, source_dir: Path) -> None:
    assert source_dir.is_dir()
    relroot = source_dir.parent.resolve()
    with zipfile.ZipFile(dest_path, "w", zipfile.ZIP_DEFLATED) as zip:
        for root, dirs, files in os.walk(source_dir):
            # add directory (needed for empty dirs)
            rel = os.path.relpath(root, relroot)
            zip.write(root, rel)
            for file in files:
                filename = os.path.join(root, file)
                if os.path.isfile(filename):  # regular files only
                    arcname = os.path.join(rel, file)
                    zip.write(filename, arcname)


def is_hidden(path: Path, hidden_list: list) -> bool:
    if path.is_dir():
        return path.parts[-1] in hidden_list
    if path.is_file():
        return path.name.startswith(".") or path.name in hidden_list
    return False


def get_filepath(path: str, base_dir: Path) -> Path:
    fp = base_dir
    if path:
        fp /= Path(unquote(path))
    return fp


def dir_exists(path: str, base_dir: Path) -> bool:
    fp = get_filepath(path, base_dir)
    return fp.exists()


def path_to_dict(i: Path, base_dir: Path, ext_dict: dict) -> FileDescriptor:
    f_name = i.name
    f_url = i.relative_to(base_dir).as_posix()
    image = get_icon(i, ext_dict)
    try:
        stat = i.stat()
        dtc = datetime.fromtimestamp(stat.st_ctime, timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        dtm = datetime.fromtimestamp(stat.st_mtime, timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        sz = "---" if i.is_dir() else size(stat.st_size)
    except:
        dtc = "---"
        dtm = "---"
        sz = "---"
    return FileDescriptor(f_name=f_name, f_url=f_url, image=image, dtc=dtc, dtm=dtm, size=sz, metadata={})


def uri_to_dict(var: str, base_dir: Path, ext_dict: dict) -> FileDescriptor:
    i = get_filepath(var, base_dir)
    f_name = i.name
    f_url = i.relative_to(base_dir).as_posix()
    image = get_icon(i, ext_dict)
    metadata = {}
    try:
        stat = i.stat()
        dtc = datetime.fromtimestamp(stat.st_ctime, timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        dtm = datetime.fromtimestamp(stat.st_mtime, timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        sz = "---" if i.is_dir() else size(stat.st_size)
    except:
        dtc = "---"
        dtm = "---"
        sz = "---"
    return FileDescriptor(f_name=f_name, f_url=f_url, image=image, dtc=dtc, dtm=dtm, size=sz, metadata=metadata)


def get_dir_listing(path: Path, base_dir: Path, ext_dict: dict, hidden_list: list):
    assert path.is_dir()
    itemList = sorted(path.iterdir(), key=lambda x: [not x.is_dir(), x])
    dir_list_dict: dict[str, FileDescriptor] = {}
    file_list_dict: dict[str, FileDescriptor] = {}

    for i in itemList:
        target = dir_list_dict if i.is_dir() else file_list_dict
        if not is_hidden(i, hidden_list):
            f_url = i.relative_to(base_dir).as_posix()
            target[f_url] = path_to_dict(i, base_dir, ext_dict)

    return dir_list_dict, file_list_dict


def run_pattern(path: str, pattern: str, mode: str, base_dir: Path) -> dict[str, str]:
    assert mode in ["name_pattern", "path_pattern"]

    # define fallback response
    metadata = {}

    # define parsers
    parse_parser = regex_parser = None
    try:
        parse_parser = parse.compile(pattern)
    except Exception as e:
        logging.warn(f"invalid pattern for 'parse' library: {pattern}", e.args)
    try:
        regex_parser = re.compile(pattern)
    except Exception as e:
        logging.warn(f"invalid pattern for 're' library: {pattern}", e.args)
    assert (parse_parser or regex_parser) is not None

    # compute the arg to parse
    path_obj = get_filepath(path, base_dir)
    if mode == "path_pattern":
        arg = path_obj.relative_to(base_dir).parent.as_posix()
        logger.info(arg)
    elif mode == "name_pattern":
        arg = path_obj.name
    else:
        raise RuntimeError()  # should never happen

    # try parsing the arg
    arg_parsed = False
    if (not arg_parsed) and (parse_parser is not None):
        try:
            result = parse_parser.parse(arg, evaluate_result=True)
            if isinstance(result, parse.Result):
                arg_parsed = True
                metadata: dict[str, str] = dict(result.named)
                logging.warn("parse_parser.parse() worked")
            else:
                logging.warn("parse_parser.parse() did not work")
        except Exception as e:
            logging.error(f"error running parse_parser on string:", e.args)
    if (not arg_parsed) and (regex_parser is not None):
        try:
            result = regex_parser.match(arg)
            if result is not None:
                arg_parsed = True
                metadata: dict[str, str] = result.groupdict()
                logging.warn("regex_parser.match() worked")
            else:
                logging.warn("regex_parse.match() did not work")
        except Exception as e:
            logging.error(f"error running regex_parser on string:", e.args)

    # return parsed metadata
    return metadata


def find_unique_attributes(selection: dict[str, FileDescriptor]) -> dict[str, list[str]]:
    uniques = defaultdict(list)
    for file, descriptor in selection.items():
        for key, value in descriptor.metadata.items():
            uniques[key].append(value)
    return dict(uniques)



import json
import logging
import os
import platform
import socket
from pathlib import Path

from flask import Flask
from flask_session import Session

from .typing import StreamSpec, FieldSpec


class Config:
    def __init__(self):
        # get host information
        self.hostname = socket.gethostname()
        self.IPAddr = socket.gethostbyname(self.hostname)
        # Config file
        template_dir = Path(__file__).parent / "templates"
        static_dir = Path(__file__).parent / "static"
        fname = "config.json"
        if platform.system() == "Windows":
            fname = "config_win.json"
        config_fp = template_dir / fname
        with open(config_fp) as json_data_file:
            data = json.load(json_data_file)
        self.hidden_list: list[str] = data["hidden"]
        self.pwd_hash: str = data["pwd_hash"]
        self.base_dir = Path(data["base_dir"]).resolve()
        self.temp_dir = Path(data["temp_dir"]).resolve()
        self.temp_dir.mkdir(exist_ok=True)
        # supported file types dict
        tp_dict_fp = template_dir / "tp_dict.json"
        with open(tp_dict_fp) as json_data_file:
            tp_dict: dict[str, dict] = json.load(json_data_file)
        self.ext_dict = {str(ext): (tp, str(data["icon"])) for tp, data in tp_dict.items() for ext in data["exts"]}
        # default settings
        self.default_spec = [
            StreamSpec(
                id="",
                name="",
                description="",
                unit="",
                frequency=0,
                fields=[FieldSpec(id="", name="", description="", dtype="f32")],
                index=[FieldSpec(id="", name="", description="", dtype="f32")]
            )
        ]
        # define app
        self.app = Flask("streaminghub-curator", template_folder=template_dir, static_folder=static_dir)
        self.app.secret_key = os.urandom(32)
        self.app.config["SESSION_TYPE"] = "filesystem"
        Session(self.app)
        self.app.logger.setLevel(logging.INFO)
        self.app.logger.info("Your Computer Name is: " + self.hostname)
        self.app.logger.info("Your Computer IP Address is: " + self.IPAddr)



import argparse
import json
from collections import defaultdict
from pathlib import Path
from urllib.parse import unquote

import numpy as np
import pandas as pd
import streaminghub_pydfds as dfds
from flask import Response, abort, jsonify, redirect, render_template, request, send_file, session
from streaminghub_datamux import gen_randseq
from werkzeug.utils import secure_filename

from . import util
from .config import Config
from .typing import FileDescriptor, StreamSpec

config = Config()


def get_selection() -> dict[str, FileDescriptor]:
    return session.get("selection", {})


def set_selection(selection: dict[str, FileDescriptor]):
    session["selection"] = selection


def get_mapping() -> dict[str, dict[str, str]] | None:
    return session.get("mapping", None)


def set_mapping(mapping: dict[str, dict[str, str]]):
    session["mapping"] = mapping


def get_stream_spec() -> list[StreamSpec]:
    return session.get("spec", config.default_spec)


def set_stream_spec(spec: list[StreamSpec]):
    session["spec"] = spec


def get_base_dir() -> Path:
    return Path(session.get("base_dir", config.base_dir))


def set_base_dir(base_dir: Path):
    session["base_dir"] = base_dir


def get_dataset_name() -> str:
    return session.get("dataset_name", "")


def set_dataset_name(dataset_name: str):
    session["dataset_name"] = dataset_name


def rereference_selection(prev_base: Path, new_base: Path):
    rel_base = new_base
    prefix = Path()
    if rel_base.is_relative_to(prev_base):
        rel_base = rel_base.relative_to(prev_base)
    else:
        prefix = prev_base.relative_to(rel_base)
    tgt: dict[str, FileDescriptor] = {}
    for k, v in get_selection().items():
        pk = Path(k)
        if pk.is_relative_to(rel_base):
            rp = pk.relative_to(rel_base)
        else:
            rp = prefix / pk
        rk = rp.as_posix()
        rv = util.uri_to_dict(rk, rel_base, config.ext_dict)
        rv.metadata = v.metadata
        tgt[rk] = rv
    return tgt


@config.app.errorhandler(400)
def err_400(error):
    return render_template("error.html", errorCode=400, errorText="Your made an invalid request."), 400


@config.app.errorhandler(403)
def err_403(error):
    return render_template("error.html", errorCode=403, errorText="The requested resource is forbidden."), 403


@config.app.errorhandler(404)
def err_404(error):
    return render_template("error.html", errorCode=404, errorText="The requested resource does not exist."), 404


@config.app.errorhandler(500)
def err_500(error):
    return render_template("error.html", errorCode=500, errorText="We ran into an internal problem"), 500


@config.app.after_request
def after_request(response: Response) -> Response:
    """ """
    response.headers.add("Accept-Ranges", "bytes")
    return response


@config.app.route("/", methods=["GET"])
def homePage():
    return redirect("/files")


@config.app.route("/files/", methods=["GET", "POST"])
@config.app.route("/files/<path:var>", methods=["GET", "POST"])
def filePage(var: str = ""):
    base_dir = Path(get_base_dir())
    path = util.get_filepath(var, base_dir)
    if not path.exists():
        abort(404)
    if not path.is_dir():
        abort(400)
    try:
        dir_dict, file_dict = util.get_dir_listing(path, base_dir, config.ext_dict, config.hidden_list)
    except:
        abort(500)

    parts = var.strip("/").split("/")
    path_html = '<a class="text-link" href="/files"><img class="icon" src="/static/root.png"></a>'
    for c in range(len(parts)):
        p_url = "/".join(parts[0 : c + 1])
        p_name = unquote(parts[c])
        path_html += f'<span class="mx-1">/</span><a class="text-link" href="/files/{p_url}">{p_name}</a>'

    dataset_name = get_dataset_name()

    # if method is POST
    config.app.logger.info(f"method={request.method}")
    if request.method == "POST":
        action = request.form.get("action")
        assert action is not None

        # update selection
        if action == "update":
            paths = request.form.getlist("selection[]")
            patch = {k: util.uri_to_dict(k, base_dir, config.ext_dict) for k in paths}
            selection = get_selection()
            selection.update(patch)
            set_selection(selection)
            config.app.logger.info(f"updated selection: {len(paths)} added")

        # drop items from selection
        if action == "drop":
            paths = request.form.getlist("selection[]")
            selection = get_selection()
            for p in paths:
                selection.pop(p, None)
            set_selection(selection)
            config.app.logger.info(f"updated selection: {len(paths)} dropped")

        # clear metadata from selection
        if action == "clear":
            paths = request.form.getlist("selection[]")
            selection = get_selection()
            for p in paths:
                selection[p].metadata.clear()
            set_selection(selection)
            config.app.logger.info(f"updated selection: {len(paths)} dropped")

        # reset selection
        if action == "reset":
            set_selection({})

        # run pattern on file name or path name
        if action in ["name_pattern", "path_pattern"]:
            selection = get_selection()
            pattern = request.form.get(action)
            assert pattern is not None
            paths = request.form.getlist("selection[]")
            for k in paths:
                assert k in selection
                uri_dict = selection[k]
                patch = util.run_pattern(k, pattern, action, base_dir)
                uri_dict.metadata.update(patch)
            set_selection(selection)

        # run dataset naming
        if action == "dataset_name":
            name = request.form.get("dataset_name")
            assert name is not None
            set_dataset_name(name)

        if action.startswith("drop_single"):
            name = action[11:]
            selection = get_selection()
            assert type(selection) == dict
            selection.pop(name, None)

        if action == "reset_root":
            if base_dir != config.base_dir:
                tgt = rereference_selection(base_dir, config.base_dir)
                set_selection(tgt)
                set_base_dir(config.base_dir)
                rp = base_dir.relative_to(config.base_dir)
                return redirect(f"/files/" + rp.as_posix())

        if action == "set_root":
            tgt = rereference_selection(base_dir, path)
            set_selection(tgt)
            set_base_dir(path)
            return redirect("/files")

        # redirection
        return redirect("/files/" + var)

    return render_template(
        "home.html",
        dataset_name=dataset_name,
        selection_dict=get_selection(),
        dir_dict=dir_dict,
        file_dict=file_dict,
        breadcrumbs=path_html,
    )


@config.app.route("/browse/<path:var>", defaults={"browse": True})
@config.app.route("/download/<path:var>", defaults={"browse": False})
def browseFile(var: str, browse: bool):
    base_dir = get_base_dir()
    path = util.get_filepath(var, base_dir)
    try:
        if not browse:
            return send_file(path, download_name=path.name)
        flag, tp, ext = util.is_media(path, config.ext_dict)
        if flag:
            return util.send_media(path, f"{tp}/{ext}")
        return send_file(path)
    except:
        abort(500)


@config.app.route("/download_dir/<path:var>")
def download_folder(var: str):
    base_dir = get_base_dir()
    path = util.get_filepath(var, base_dir)
    assert path.is_dir()
    if util.is_hidden(path, config.hidden_list):
        abort(403)
    zip_name = path.with_suffix(".zip").name
    zip_path = config.temp_dir / zip_name
    try:
        util.zip_directory(zip_path, path)
        return send_file(zip_path, download_name=zip_name)
    except:
        abort(500)


@config.app.route("/upload/", methods=["POST"])
@config.app.route("/upload/<path:var>", methods=["POST"])
def upload_file(var: str = ""):
    base_dir = get_base_dir()
    text = ""
    path = util.get_filepath(var, base_dir)
    assert path.is_dir()
    if util.is_hidden(path, config.hidden_list):
        abort(403)
    files = request.files.getlist("files[]")
    num_total = len(files)
    num_failed = 0
    for file in files:
        assert file.filename
        filename = secure_filename(file.filename)
        fupload = path / filename
        if not fupload.exists():
            try:
                file.save(fupload)
                config.app.logger.info(filename + " Uploaded")
                text = text + filename + " Uploaded<br>"
            except Exception as e:
                num_failed += 1
                config.app.logger.info(filename + " Failed with Exception " + str(e))
                text = text + filename + " Failed with Exception " + str(e) + "<br>"
                continue
        else:
            config.app.logger.info(filename + " Failed because File Already Exists or File Type Issue")
            text = text + filename + " Failed because File Already Exists or File Type not secure <br>"
    return render_template("uploadsuccess.html", text=text, fileNo=num_total, fileNo2=num_failed)


@config.app.route("/streams", methods=["GET"])
def get_streams():
    selection = get_selection()
    uniques = util.find_unique_attributes(selection)
    pattern = list(uniques.keys())
    return render_template("streams.html", file_dict=selection, pattern=pattern, location="")


@config.app.route("/metadata", methods=["POST"])
def get_metadata():
    files = request.get_json()["files"]
    selection = get_selection()
    metadata = {}
    for file in files:
        if file not in selection:
            abort(400)
        metadata[file] = selection[file].metadata
    return metadata


@config.app.route("/export", methods=["GET"])
def get_export():
    selection = get_selection()
    uniques = util.find_unique_attributes(selection)
    pattern = list(uniques.keys())
    return render_template("export.html", file_dict=selection, pattern=pattern, location="")


@config.app.route("/export", methods=["POST"])
def do_export():

    # get saved metadata and form input
    dataset_name = get_dataset_name()
    selection = get_selection()
    mapping = get_mapping()  # stream_id: {field_id: column_name}
    spec = get_stream_spec()
    base_dir = get_base_dir()
    format = str(request.form.get("format"))
    pattern = request.form.getlist("pattern[]")
    pattern_string = f"{format}://" + "_".join(["{" + k + "}" for k in pattern])

    config.app.logger.info(f"format={format}, pattern={pattern}")
    assert mapping is not None
    assert format in ["h5", "csv", "parquet", "npy"]
    assert len(pattern) > 0

    # create temporary directory for dataset
    dataset_dir = config.temp_dir / gen_randseq(20)
    dataset_dir.mkdir(exist_ok=True, parents=True)

    data_dir = dataset_dir / dataset_name
    data_dir.mkdir(exist_ok=True)

    groups: dict[str, set[str]] = {}
    for path, descriptor in selection.items():
        # TODO modify hardcoded "root" value for HDF5 to work
        _, data = dfds.create_reader(base_dir / path).read("root")
        # replace meta with stuff from "descriptor"
        meta = descriptor.metadata
        # update groups information
        for k, v in meta.items():
            items = groups.get(k, set())
            items.add(v)
            groups[k] = items
        # get data ("data") columns ("src")
        # and save them ("stream_data") with new names ("tgt")
        stream_data = {}
        for stream_id, colmap in mapping.items():
            for field, col in colmap.items():
                stream_data[field] = data[col].to_numpy()
        # generate destination df and path
        dest_df = pd.DataFrame(stream_data)
        dest_fp = data_dir / str("_".join([meta[k] for k in pattern]) + f".{format}")
        # save data to path
        if format == "h5":
            raise NotImplementedError()
        if format == "csv":
            dest_df.to_csv(dest_fp, index=False)
        if format == "parquet":
            dest_df.to_parquet(dest_fp, index=False)
        if format == "npy":
            np.save(dest_fp, dest_df.to_records(index=False))
        config.app.logger.debug(f"{path} -> {dest_fp} (cols={dest_df.columns.tolist()})")

    # generate collection metadata
    collection = dfds.Collection(
        name=dataset_name,
        description="Generated by StreamingHub Curator",
        keywords=[],
        authors=[],
        streams={
            s.id: dfds.Stream(
                name=s.name,
                description=s.description,
                unit=s.unit,
                frequency=s.frequency,
                fields={f.id: dfds.Field(**f.model_dump()) for f in s.fields},
                index={f.id: dfds.Field(**f.model_dump()) for f in s.index},
            )
            for s in spec
        },
        groups={k: dfds.Group(description="", values=sorted(v)) for k, v in groups.items()},
        pattern=pattern_string,
    )
    with open(dataset_dir / f"{dataset_name}.collection.json", "w") as file:
        json.dump(collection.model_dump(), file, indent=2)

    # compress dataset into tar file
    tar_path = dataset_dir / f"{dataset_name}.tar.gz"
    util.make_tarfile(tar_path, dataset_dir, dataset_name)

    return send_file(tar_path, download_name=tar_path.name)


@config.app.route("/streamspec", methods=["GET"])
def onreq_stream_spec():
    spec = get_stream_spec()
    return jsonify([s.model_dump() for s in spec])


@config.app.route("/streamspec", methods=["POST"])
def update_stream_spec():
    spec: list[dict] = request.get_json()
    config.app.logger.info(f"spec={spec}")
    set_stream_spec([StreamSpec(**s) for s in spec])
    return jsonify(success=True, error=None)


@config.app.route("/attributes", methods=["GET"])
def get_attribute_editor():
    return render_template("attributes.html")


@config.app.route("/attribmap", methods=["GET"])
def get_attribute_map():
    # forward index
    # { path, ref } -> { {col, dtype}, meta }
    fwd_index: dict[tuple[str, str], dict[str, dict[str, str]]] = {}
    base_dir = get_base_dir()
    for path, desc in get_selection().items():
        cols = dfds.create_reader(base_dir / path).lsfields()
        meta = desc.metadata
        # broadcast metadata to all entries in the fields dict
        for ref, colmap in cols.items():
            fwd_index[(path, ref)] = dict(cols=colmap, meta=meta)

    # inverse index
    # {col} -> [ {path, ref, dtype, meta} ]
    inv_index: defaultdict[str, list[dict[str, str | dict[str, str]]]] = defaultdict(list)
    for (path, ref), entries in fwd_index.items():
        cols, meta = entries["cols"], entries["meta"]
        for col, dtype in cols.items():
            inv_index[col].append(
                dict(
                    path=path,
                    ref=ref,
                    dtype=dtype,
                    meta=meta,
                )
            )

    # number of units in selection
    file_count = len(fwd_index)
    cols_count = {str(k): len(v) for k, v in inv_index.items()}
    # get max occurence count of the attributes
    max_count = max(cols_count.values())
    # ensure at least one column is common across all files
    assert max_count == file_count
    final_index = {
        k: {
            "items": v,
            "summary": {
                "count": len(v),
                # bool indicating if attribute has full coverage
                "fully_covered": len(v) == max_count,
            },
        }
        for k, v in inv_index.items()
    }
    return jsonify(cols=final_index, file_count=file_count, mapping=get_mapping())


@config.app.route("/attribmap", methods=["POST"])
def set_attribute_editor():
    mapping = request.get_json()
    config.app.logger.info(f"mapping={mapping}")
    set_mapping(mapping)
    return jsonify(success=True, error=None)


if __name__ == "__main__":

    parser = argparse.ArgumentParser("Streaminghub Curator")
    parser.add_argument("--host", "-H", type=str, default="0.0.0.0")
    parser.add_argument("--port", "-p", type=int, default=8000)
    args = parser.parse_args()

    config.app.run(host=args.host, port=args.port, debug=True)
